/*!
 * \file      smtc_hal_adc.c
 *
 * \brief     ADC Hardware Abstraction Layer implementation
 *
 * The Clear BSD License
 * Copyright Semtech Corporation 2021. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted (subject to the limitations in the disclaimer
 * below) provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Semtech corporation nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
 * THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL SEMTECH CORPORATION BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * -----------------------------------------------------------------------------
 * --- DEPENDENCIES ------------------------------------------------------------
 */
#include <stdint.h>   // C99 types
#include <stdbool.h>  // bool type

#include "smtc_hal.h"
#include "modem_pinout.h"

#include "system_FM33A0XXEV.h"
#include "fm33a0xxev_include_all.h"

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE MACROS-----------------------------------------------------------
 */
#define const_adc_Slope ( *( ( uint16_t* ) ( 0x1FFFFA84 ) ) )  // ADCбʣ1000ʹ
#define const_adc_Offset ( *( ( int16_t* ) ( 0x1FFFFA86 ) ) )  // ADCؾ࣬100ʹãλmV
#define const_T_30 ( *( ( uint16_t* ) ( 0x1FFFFA92 ) ) )       // 0X640  30ֵ

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE CONSTANTS -------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE TYPES -----------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE VARIABLES -------------------------------------------------------
 */

float const_TmpeK  = 4.5295;  // 0X640 ADCб
float const_30_top = 30.0;

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE FUNCTIONS DECLARATION -------------------------------------------
 */

static uint16_t ANAC_ADC_VoltageCalc( uint32_t ADCData );
static float    ANAC_ADC_TemperatureCalc( uint32_t ADCData );
static uint8_t  ADC_Wait_Finish( void );

/*
 * -----------------------------------------------------------------------------
 * --- PUBLIC FUNCTIONS DEFINITION ---------------------------------------------
 */

void hal_adc_init( void )
{
    CMU_PERCLK_SetableEx( CICCLK, ENABLE );  //ʹADCʱ
    CDIF_CR_INTF_EN_Setable( ENABLE );       //Դӿʹ
    VRTC_Init_RCMF_Trim( );
    VRTC_RCMFCR_EN_Setable( ENABLE );
    VRTC_ADCCR_CKS_Set( VRTC_ADCCR_CKS_RCMF_4 );     // ADCʱѡ--500K
    VRTC_ADCCR_CKE_Setable( ENABLE );                // ADCʱʹ
    ADC_CFGR_BUFSEL_Set( ADC_CFGR_BUFSEL_ADC_IN1 );  // ADCͨѡ

    ADC_CFGR_BUFEN_Setable( ENABLE );      // ADCͨbufferʹ/ֹ
    ADC_CR_MODE_Set( ADC_CR_MODE_INNER );  // ADCģʽѡڲۼ

    ADC_TRIM_Write( 0X3FF );  // adcƵ500Kʱʱ4ms

    ADC_CR_ADC_IE_Setable( DISABLE );  //ڲۼģʽжϽֹ
    ADC_CR_EN_Setable( DISABLE );      // ADCر
}

uint16_t hal_adc_get_vbat( void )
{
    uint16_t          ADCData[8];
    int8_t            i;
    volatile uint16_t millivolt = 0;
    uint32_t          fTempADC  = 0;

    ADC_CFGR_BUFSEL_Set( ADC_CFGR_BUFSEL_VBAT );  // ADCͨѡ--VBAT
    ADC_CR_EN_Setable( ENABLE );                  // ADC

    fTempADC = 0;
    for( i = 0; i < 8; i++ )
    {
        ADC_ISR_ADC_IF_Clr( );  //жϱ־

        ADCData[i] = 0;
        if( 0 == ADC_Wait_Finish( ) )  //ȴת
        {
            ADCData[i] = ADC_DR_Read( );  //ȡADֵ
            fTempADC += ADCData[i];
        }
        else
        {
            break;
        }
    }

    if( i == 8 )
    {
        fTempADC = fTempADC / 8;
    }

    millivolt = ANAC_ADC_VoltageCalc( fTempADC );  // ADֵתΪѹ,ԴѹΪ5V

    ADC_CR_EN_Setable( DISABLE );  // ADCر

    return millivolt;
}

int8_t hal_adc_get_temp( void )
{
    uint16 ADCData[8];
    uint08 i;
    float  Temperature = 0;
    uint32 fTempADC    = 0;

    ADC_CFGR_BUFSEL_Set( ADC_CFGR_BUFSEL_TS );  // ADCͨѡ
    ADC_TRIM_Write( 0X640 );                    //
    ADC_CR_EN_Setable( ENABLE );                // ADC

    fTempADC = 0;
    for( i = 0; i < 8; i++ )
    {
        ADC_ISR_ADC_IF_Clr( );  //жϱ־

        ADCData[i] = 0;
        if( 0 == ADC_Wait_Finish( ) )  //ȴת
        {
            ADCData[i] = ADC_DR_Read( );  //ȡADֵ
            fTempADC += ADCData[i];
        }
        else
        {
            break;
        }
    }

    if( i == 8 )
    {
        fTempADC = fTempADC / 8;
    }

    Temperature = ANAC_ADC_TemperatureCalc( fTempADC );

    ADC_CR_EN_Setable( DISABLE );  // ADCر

    return Temperature;
}

void hal_adc_deinit( void )
{
    ADC_CR_EN_Setable( DISABLE );       // ADCر
    ADC_CFGR_BUFEN_Setable( DISABLE );  // ADCͨbufferʹ/ֹ
    VRTC_ADCCR_CKE_Setable( DISABLE );  // ADCʱӹر
}

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE FUNCTIONS DEFINITION --------------------------------------------
 */

uint16_t ANAC_ADC_VoltageCalc( uint32_t ADCData )
{
    int32_t Volt = 0;
    Volt         = ( ADCData * const_adc_Slope / 1000 ) + const_adc_Offset / 100;

    if( Volt & 0x80000000 )  // VoltΪֵ
    {
        Volt = 0;
    }
    return Volt;
}

float ANAC_ADC_TemperatureCalc( uint32_t ADCData )
{
    float T = 0;
    if( ADCData > const_T_30 )
    {
        T = ( ADCData - const_T_30 ) / const_TmpeK + const_30_top;
    }
    else
    {
        T = const_30_top - ( const_T_30 - ADCData ) / const_TmpeK;
    }
    return T;
}

uint8_t ADC_Wait_Finish( void )
{
    uint32 timeout = 0;
    do
    {
        if( SET == ADC_ISR_ADC_IF_Chk( ) )
            return 0;

    } while( timeout++ < 0xFFFFFFFFU );
    return 1;  //ʱ
}
