/*!
 * \file      smtc_hal_flash.c
 *
 * \brief     FLASH Hardware Abstraction Layer implementation
 *
 * The Clear BSD License
 * Copyright Semtech Corporation 2021. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted (subject to the limitations in the disclaimer
 * below) provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Semtech corporation nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
 * THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL SEMTECH CORPORATION BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * -----------------------------------------------------------------------------
 * --- DEPENDENCIES ------------------------------------------------------------
 */

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "smtc_hal.h"
#include "system_FM33A0XXEV.h"
#include "fm33a0xxev_include_all.h"
/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE MACROS-----------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE CONSTANTS -------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE TYPES -----------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE VARIABLES -------------------------------------------------------
 */

uint32_t flash_user_start_addr = FLASH_USER_END_ADDR;

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE FUNCTIONS DECLARATION -------------------------------------------
 */

static bool FlashSectorErase( uint32_t address );
static bool programMultiWords( uint32_t addr, const uint32_t* data, uint32_t size );

/*
 * -----------------------------------------------------------------------------
 * --- PUBLIC FUNCTIONS DEFINITION ---------------------------------------------
 */

uint8_t hal_flash_erase_page( uint32_t addr, uint8_t nb_page )
{
    for( uint8_t page = 0; page < nb_page; page++ )
    {
        if( !FlashSectorErase( addr + page * ADDR_FLASH_PAGE_SIZE ) )
            return false;
    }

    return true;
}

uint32_t hal_flash_write_buffer( uint32_t addr, const uint8_t* buffer, uint32_t size )
{
    ( void ) programMultiWords( addr, ( uint32_t* ) buffer, ( ( size + 3U ) / 4U ) );
    return 0;
}

void hal_flash_read_buffer( uint32_t addr, uint8_t* buffer, uint32_t size )
{
    while( size-- )
    {
        *buffer = *( uint8_t* ) addr & 0xff;
        buffer += 1;
        addr += 1;
    }
}

uint32_t hal_flash_get_user_start_addr( void )
{
    return flash_user_start_addr;
}

void hal_flash_set_user_start_addr( uint32_t addr )
{
    flash_user_start_addr = addr;
}

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE FUNCTIONS DEFINITION --------------------------------------------
 */

bool FlashSectorErase( uint32_t address )  // actually it operate SECTOR,SIZE= 2KB!!!!!!
{
#define FLS_SECTOR_ERASE_KEY0 0x96969696
#define FLS_SECTOR_ERASE_KEY1 0x3C3C3C3C
#define FLS_SECTOR_ERASE_DATA 0x1234ABCD

    uint16_t i;
    CMU_PERCLK_SetableEx( FLASHCLK, ENABLE );
    CMU_OPCCR2_NVMCKE_Setable( ENABLE );

    FLS_EPCR_ERTYPE_Set( FLS_EPCR_ERTYPE_SECTOR );  //ѡ--

    FLS_EPCR_EREQ_Set( FLS_EPCR_EREQ_Msk );
    FLS_KEY_Write( FLS_SECTOR_ERASE_KEY0 );
    FLS_KEY_Write( FLS_SECTOR_ERASE_KEY1 );
    *( uint32_t* ) address = FLS_SECTOR_ERASE_DATA;

    while( SET != FLS_ISR_ERD_Chk( ) )
    {
        __NOP( );
    }
    FLS_ISR_ERD_Clr( );

    FLS_KEY_Write( 0x00000000 );

    CMU_OPCCR2_NVMCKE_Setable( DISABLE );
    CMU_PERCLK_SetableEx( FLASHCLK, DISABLE );
    for( i = 0; i < 128; i++ )
    {
        if( *( uint32_t* ) ( address + i * 4 ) != 0xFFFFFFFF )
        {
            return false;
        }
    }
    return true;

#undef FLS_SECTOR_ERASE_KEY0
#undef FLS_SECTOR_ERASE_KEY1
#undef FLS_SECTOR_ERASE_DATA
}

bool programMultiWords( uint32_t addr, const uint32_t* data, uint32_t size )
{
#define FLS_PROG_KEY0 0xA5A5A5A5
#define FLS_PROG_KEY1 0xF1F1F1F1

    uint32_t primask;

    CMU_PERCLK_SetableEx( FLASHCLK, ENABLE );
    CMU_OPCCR2_NVMCKE_Setable( ENABLE );

    FLS_ISR_KEYERR_Clr( );
    FLS_EPCR_PREQ_Set( FLS_EPCR_PREQ_Msk );
    primask = __get_PRIMASK( );
    __disable_irq( );
    FLS_KEY_Write( FLS_PROG_KEY0 );  // дflashkey0
    FLS_KEY_Write( FLS_PROG_KEY1 );  // дflashkey1
    __set_PRIMASK( primask );

    while( size-- )
    {
        FLS_EPCR_PREQ_Set( FLS_EPCR_PREQ_Msk );
        *( uint32_t* ) addr = *data;
        data += 1;
        addr += 4;
        while( FLS_ISR_PRD_Chk( ) == RESET )
            ;                // ȴд
        FLS_ISR_PRD_Clr( );  // джϱ־
    }

    FLS_KEY_Write( 0x00000000 );

    CMU_OPCCR2_NVMCKE_Setable( DISABLE );
    CMU_PERCLK_SetableEx( FLASHCLK, DISABLE );

#undef FLS_PROG_KEY0
#undef FLS_PROG_KEY1

    return 0;
}

/* --- EOF ------------------------------------------------------------------ */
