/*!
 * \file      smtc_hal_gpio.c
 *
 * \brief     GPIO Hardware Abstraction Layer implementation
 *
 * The Clear BSD License
 * Copyright Semtech Corporation 2021. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted (subject to the limitations in the disclaimer
 * below) provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Semtech corporation nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
 * THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL SEMTECH CORPORATION BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * -----------------------------------------------------------------------------
 * --- DEPENDENCIES ------------------------------------------------------------
 */
#include <stdint.h>   // C99 types
#include <stdbool.h>  // bool type
#include "smtc_hal.h"
#include "core_cm0plus.h"
#include "system_FM33A0XXEV.h"
#include "fm33a0xxev_include_all.h"

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE MACROS-----------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE CONSTANTS -------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE TYPES -----------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE VARIABLES -------------------------------------------------------
 */

/*!
 * Array holding attached IRQ gpio data context
 */

static hal_gpio_irq_t const* gpio_irq[24];

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE FUNCTIONS DECLARATION -------------------------------------------
 */

static void CloseIO( GPIO_Type* GPIOx, uint32_t GPIO_Pin );

//żж
static void get_pin_shift_irq_line( hal_gpio_pin_names_t pin, uint8_t* irq_line );

/*
 * -----------------------------------------------------------------------------
 * --- PUBLIC FUNCTIONS DEFINITION ---------------------------------------------
 */

void hal_gpio_init_out( const hal_gpio_pin_names_t pin, const uint32_t value )
{
    GPIO_Type* gpio_port = ( GPIO_Type* ) ( GPIOA_BASE + ( pin / 16 ) * 0x40 );
    uint32_t   gpio_pin  = ( uint32_t )( 1 << ( pin & 0xF ) );
    CMU_PERCLK_SetableEx( PADCLK, ENABLE );  // IO CLK

    GPIO_InitTypeDef gpio_init_struct;
    gpio_init_struct.Pin    = gpio_pin;
    gpio_init_struct.PxFCR  = GPIO_FCR_OUT;
    gpio_init_struct.PxINEN = GPIO_IN_Dis;
    gpio_init_struct.PxODEN = GPIO_OD_Dis;
    gpio_init_struct.PxPUEN = GPIO_PU_Dis;
    GPIO_Init( gpio_port, &gpio_init_struct );

    if( value )
        GPIO_SetBits( gpio_port, gpio_pin );
    else
        GPIO_ResetBits( gpio_port, gpio_pin );
}

void hal_gpio_deinit( const hal_gpio_pin_names_t pin )
{
    GPIO_Type* gpio_port = ( GPIO_Type* ) ( GPIOA_BASE + ( pin / 16 ) * 0x40 );
    uint32_t   gpio_pin  = ( uint32_t )( 1 << ( pin & 0xF ) );

    CloseIO( gpio_port, gpio_pin );
}

void hal_gpio_set_value( const hal_gpio_pin_names_t pin, const uint32_t value )
{
    GPIO_Type* gpio_port = ( GPIO_Type* ) ( GPIOA_BASE + ( pin / 16 ) * 0x40 );
    uint32_t   gpio_pin  = ( uint32_t )( 1 << ( pin & 0xF ) );

    if( value )
        GPIO_SetBits( gpio_port, gpio_pin );
    else
        GPIO_ResetBits( gpio_port, gpio_pin );
}

uint32_t hal_gpio_get_value( const hal_gpio_pin_names_t pin )
{
    GPIO_Type* gpio_port = ( GPIO_Type* ) ( GPIOA_BASE + ( pin / 16 ) * 0x40 );
    uint32_t   gpio_pin  = ( uint32_t )( 1 << ( pin & 0xF ) );

    return GPIO_ReadInputDataBit( gpio_port, gpio_pin );
}

void hal_gpio_toggle( const hal_gpio_pin_names_t pin )
{
    GPIO_Type* gpio_port = ( GPIO_Type* ) ( GPIOA_BASE + ( pin / 16 ) * 0x40 );
    uint32_t   gpio_pin  = ( uint32_t )( 1 << ( pin & 0xF ) );
    GPIO_ToggleBits( gpio_port, gpio_pin );
}

void hal_gpio_init_in( const hal_gpio_pin_names_t pin, const hal_gpio_pull_mode_t pull_mode,
                       const hal_gpio_irq_mode_t irq_mode, hal_gpio_irq_t* irq )
{
    const uint32_t irq_modes[] = { EXTIDISABLE, EXTI_RISING, EXTI_FALLING, EXTI_BOTH };

    GPIO_Type* gpio_port = ( GPIO_Type* ) ( GPIOA_BASE + ( pin / 16 ) * 0x40 );
    uint32_t   gpio_pin  = ( uint32_t )( 1 << ( pin & 0x0F ) );

    CMU_PERCLK_SetableEx( PADCLK, ENABLE );  // IO CLK

    GPIO_InitTypeDef gpio_init_struct;
    gpio_init_struct.Pin    = gpio_pin;
    gpio_init_struct.PxFCR  = GPIO_FCR_IN;
    gpio_init_struct.PxINEN = GPIO_IN_En;
    gpio_init_struct.PxODEN = GPIO_OD_Dis;
    if( pull_mode == BSP_GPIO_PULL_MODE_UP )
    {
        gpio_init_struct.PxPUEN = GPIO_PU_En;
    }
    else
    {
        gpio_init_struct.PxPUEN = GPIO_PU_Dis;
    }

    GPIO_Init( gpio_port, &gpio_init_struct );

    if( irq != NULL )
    {
        irq->pin = pin;
    }

    if( irq_modes[irq_mode] != EXTIDISABLE )
    {
        CMU_OPCCR1_EXTICKSEL_Set( CMU_OPCCR1_EXTICKSEL_LSCLK );  // Use APBLCK to sample exti
        CMU_OPCCR1_EXTICKE_Setable( ENABLE );
        hal_mcu_delay_ms( 1 );  // Wait at least 4 APBCLK
        GPIO_EXTI_Init( gpio_port, gpio_pin, irq_modes[irq_mode], DISABLE );

        hal_gpio_irq_attach( irq );

        // NVIC Configuration
        hal_gpio_irq_disable( );
        NVIC_SetPriority( GPIO_IRQn, 1 );
        hal_gpio_irq_enable( );
    }
}

void hal_gpio_irq_attach( const hal_gpio_irq_t* irq )
{
    uint8_t irq_line = 0;

    if( ( irq != NULL ) && ( irq->callback != NULL ) )
    {
        get_pin_shift_irq_line( irq->pin, &irq_line );

        gpio_irq[irq_line] = irq;
    }
}

void hal_gpio_irq_deatach( const hal_gpio_irq_t* irq )
{
    uint8_t irq_line = 0;

    if( irq != NULL )
    {
        get_pin_shift_irq_line( irq->pin, &irq_line );

        gpio_irq[irq_line] = NULL;
    }
}

void hal_gpio_irq_enable( void )
{
    NVIC_EnableIRQ( GPIO_IRQn );
}

void hal_gpio_irq_disable( void )
{
    NVIC_DisableIRQ( GPIO_IRQn );
}

bool hal_gpio_is_pending_irq( const hal_gpio_pin_names_t pin )
{
    GPIO_Type* gpio_port = ( GPIO_Type* ) ( GPIOA_BASE + ( pin / 16 ) * 0x40 );
    uint32_t   gpio_pin  = ( uint32_t )( 1 << ( pin & 0xF ) );
    return ( bool ) GPIO_EXTI_EXTIISR_ChkEx( gpio_port, gpio_pin );
}

void hal_gpio_clear_pending_irq( const hal_gpio_pin_names_t pin )
{
    NVIC_ClearPendingIRQ( GPIO_IRQn );
}

/**
 * @brief This function handles EXTI line0 interrupt.
 */
void GPIO_IRQHandler( void )
{
    GPIO_Type* gpio_port = NULL;
    uint32_t   gpio_pin  = 0;

    for( uint8_t i = 0; i < 24; i++ )
    {
        if( gpio_irq[i] != NULL )
        {
            gpio_port = ( GPIO_Type* ) ( GPIOA_BASE + ( gpio_irq[i]->pin / 16 ) * 0x40 );
            gpio_pin  = ( uint32_t )( 1 << ( gpio_irq[i]->pin & 0xF ) );
            if( SET == GPIO_EXTI_EXTIISR_ChkEx( gpio_port, gpio_pin ) )
            {
                GPIO_EXTI_EXTIISR_ClrEx( gpio_port, gpio_pin );
                if( gpio_irq[i]->callback != NULL )
                {
                    gpio_irq[i]->callback( gpio_irq[i]->context );
                }
                break;
            }
        }
    }
}

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE FUNCTIONS DEFINITION --------------------------------------------
 */

// IOرգGPIOA,B,C,D,E,F,Godߣ
//IOΪ̬
static void CloseIO( GPIO_Type* GPIOx, uint32_t GPIO_Pin )
{
    uint32_t pinpos = 0x00, pinbit = 0x00;

    for( pinpos = 0; pinpos < 16; pinpos++ )
    {
        pinbit = ( ( uint32_t ) 0x01 ) << pinpos;
        if( GPIO_Pin & pinbit )
        {
            GPIOx->INEN &= ~( 1 << pinpos );
            GPIOx->FCR &= ~( 3 << ( pinpos * 2 ) );
        }
    }
}

static void get_pin_shift_irq_line( hal_gpio_pin_names_t pin, uint8_t* irq_line )
{
    uint32_t pin_num = 0;
    uint32_t i;
    uint32_t gpio_base = 0;
    uint32_t gpio_pin  = 0;

    gpio_base = GPIOA_BASE + ( pin / 16 ) * 0x40;
    gpio_pin  = ( uint32_t )( 1 << ( pin & 0xF ) );

    for( i = 0; i < 16; i++ )
    {
        if( gpio_pin & ( 1 << i ) )
        {
            pin_num = i;
            break;
        }
    }

    if( i == 16 )
        return;

    switch( gpio_base )
    {
    case GPIOA_BASE:
    case GPIOB_BASE:
    case GPIOC_BASE:
    case GPIOD_BASE:
    case GPIOE_BASE:
    case GPIOF_BASE:
        if( ( gpio_base == GPIOE_BASE ) || ( gpio_base == GPIOF_BASE ) )
        {
            if( pin_num > 11 )
                break;
        }
        else
        {
            if( pin_num > 15 )
                break;
        }

        if( ( pin_num <= 3 ) )
        {
            *irq_line = 0;
        }
        else if( ( pin_num >= 4 ) && ( pin_num <= 7 ) )
        {
            *irq_line = 1;
        }
        else if( ( pin_num >= 8 ) && ( pin_num <= 11 ) )
        {
            *irq_line = 2;
        }
        else
        {
            *irq_line = 3;
        }

        if( gpio_base == GPIOB_BASE )
        {
            *irq_line += 4;
        }
        if( gpio_base == GPIOC_BASE )
        {
            *irq_line += 8;
        }
        if( gpio_base == GPIOD_BASE )
        {
            *irq_line += 12;
        }
        if( gpio_base == GPIOE_BASE )
        {
            *irq_line += 16;
        }
        if( gpio_base == GPIOF_BASE )
        {
            *irq_line += 20;
        }
        break;
    default:
        break;
    }
}

/* --- EOF ------------------------------------------------------------------ */
