/*!
 * @file      smtc_hal_lp_timer.c
 *
 * @brief     RTC Hardware Abstraction Layer implementation
 *
 * The Clear BSD License
 * Copyright Semtech Corporation 2023. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted (subject to the limitations in the disclaimer
 * below) provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Semtech corporation nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
 * THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL SEMTECH CORPORATION BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * -----------------------------------------------------------------------------
 * --- DEPENDENCIES ------------------------------------------------------------
 */

#include "smtc_hal.h"
#include "system_FM33A0XXEV.h"
#include "fm33a0xxev_include_all.h"

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE MACROS-----------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE CONSTANTS -------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE TYPES -----------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE VARIABLES -------------------------------------------------------
 */

/* ڸ±ļֵ */
static uint32_t lptim_timing_stored_time_prev;
static uint32_t lptim_timing_system_time;

static volatile bool wut_timer_irq_happened = false;

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE FUNCTIONS DECLARATION -------------------------------------------
 */

/**
 * @brief Read the 32-bit timer hardware time
 *
 * @returns The current time value hardware time
 */
static uint32_t lptim_get_hardware_time( void );

/**
 * @brief Update the count value
 * @lptim_timing_stored_time_prev, @lptim_timing_system_time
 */
static void lptim_timing_update_system_time( uint32_t hardware_time );

/**
 * @brief Get the count value
 */
static uint32_t lptim_timing_get_system_time( void );

/*!
 * @brief Converts time in ms to time in wake up timer ticks
 *
 * @param [in] milliseconds Time in milliseconds
 * @returns ticks Time in wake up timer ticks
 */
static uint32_t hal_lptim_ms_2_wakeup_timer_tick( const int32_t milliseconds );

/*
 * -----------------------------------------------------------------------------
 * --- PUBLIC FUNCTIONS DEFINITION ---------------------------------------------
 */

void hal_lp_timer_init( void )
{
    CMU_PERCLK_SetableEx( LPTCLK, ENABLE );            //ʱʹ
    CMU_OPCCR2_LPTCKS_Set( CMU_OPCCR2_LPTCKS_LSCLK );  //ѡʱԴ
    CMU_OPCCR2_LPTCKE_Setable( ENABLE );               //ʱԴʹ

    LPTIM_CFGR_ETR_AFEN_Setable( DISABLE );             // ETR˲
    LPTIM_CFGR_PSCSEL_Set( LPTIM_CFGR_PSCSEL_CLKSEL );  //ʱѡ
    LPTIM_CFGR_DIVSEL_Set( LPTIM_CFGR_DIVSEL_DIV4 );    //ʱӷƵѡ
    LPTIM_CFGR_ONST_Set( LPTIM_CFGR_ONST_CONTINUE );    //ֹͣ
    LPTIM_CFGR_TMODE_Set( LPTIM_CFGR_TMODE_COUNTER );   //ͨʱģʽ

    LPTIM_ARR_Write( 0xFFFFFFFF );     // PWM
    LPTIM_ISR_OVIF_Clr( );             /* жϱ־λ */
    LPTIM_IER_OVIE_Setable( DISABLE ); /* ж */

    // CCRʱж
    LPTIM_CCSR_CC1S_Set( LPTIM_CCSR_CC1S_LPT_CH1_OUT );  // LPT32_CH1ȽϹ,Ҫ
    LPTIM_CCSR_POLAR1_Set( LPTIM_CCSR_POLAR1_POS );      // LPT32_CH1μ
    LPTIM_CCR1_Write( 0xFFFFFFFF );                      // LPT32_CH1 PWMռձ
    LPTIM_ISR_CC1IF_Clr( );                              //жϱ־
    LPTIM_IER_CC1IE_Setable( ENABLE );                   /* ͨ1ж */

    NVIC_DisableIRQ( LPTIM_IRQn );
    NVIC_SetPriority( LPTIM_IRQn, 2 );  //жȼ
    NVIC_EnableIRQ( LPTIM_IRQn );

    LPTIM_CR_EN_Setable( ENABLE ); /* ʹܼ: */

    lptim_timing_stored_time_prev = 0;
    lptim_timing_system_time      = 0;
}

void LPTIM_IRQHandler( void )  // PWMжϷ
{
    /**************жλ־*********************/
    if( LPTIM_ISR_CC1IF_Chk( ) == SET )  //жϱ־
    {
        LPTIM_ISR_CC1IF_Clr( );  //жϱ־
        wut_timer_irq_happened = true;
        LPTIM_IER_CC1IE_Setable( DISABLE ); /* رͨ1ж */
    }
}

uint32_t hal_lp_timer_get_time_s( void )
{
    uint32_t time;

    time = ( uint32_t ) lptim_timing_get_system_time( );
    /* 32.768KHzʱ4Ƶ󣬵8192 = 2^13 */
    time = time >> 13;

    return time;
}

uint32_t hal_lp_timer_get_time_ms( void )
{
    uint64_t time;

    time = ( ( uint64_t ) lptim_timing_get_system_time( ) ) * 1000;
    /* 32.768KHzʱ4Ƶ󣬵8192 = 2^13 */
    time = time >> 13;

    return ( uint32_t ) time;
}

uint32_t hal_lp_timer_get_time_100us( void )
{
    uint64_t time;

    time = ( ( uint64_t ) lptim_timing_get_system_time( ) ) * 10000;
    /* 32.768KHzʱ4Ƶ󣬵8192 = 2^13 */
    time = time >> 13;

    return ( uint32_t ) time;
}

void hal_lptim_wakeup_timer_set_ms( const int32_t milliseconds )
{
    uint32_t delay_ms_2_tick;
    uint32_t current_tick;

    /* reset irq status */
    wut_timer_irq_happened = false;

    delay_ms_2_tick = hal_lptim_ms_2_wakeup_timer_tick( milliseconds );
    current_tick    = lptim_get_hardware_time( );

    LPTIM_ISR_CC1IF_Clr( );                              //жϱ־
    LPTIM_CCR1_Write( delay_ms_2_tick + current_tick );  // LPT32_CH1 PWMռձ
    LPTIM_IER_CC1IE_Setable( ENABLE );                   /* ͨ1ж */
}

bool hal_lptim_has_wut_irq_happened( void )
{
    return wut_timer_irq_happened;
}

void hal_lptim_wakeup_timer_stop( void )
{
    LPTIM_IER_CC1IE_Setable( DISABLE ); /* رͨ1ж */
}

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE FUNCTIONS DEFINITION --------------------------------------------
 */

static uint32_t hal_lptim_ms_2_wakeup_timer_tick( const int32_t milliseconds )
{
    uint32_t tick;

    /* 32.768KHzʱ4Ƶ󣬵8192 = 2^13 */
    tick = ( ( uint32_t ) milliseconds ) << 13;
    tick = tick / 1000;
    return tick;
}

static uint32_t lptim_get_hardware_time( void )
{
    uint32_t t_old, t;

    t = LPTIM->CNTR;
    do
    {
        t_old = t;
        t     = LPTIM->CNTR;
    } while( t != t_old );

    return t;
}

static void lptim_timing_update_system_time( uint32_t hardware_time )
{
    uint32_t delta;

    CRITICAL_SECTION_BEGIN( );
    delta                         = hardware_time - lptim_timing_stored_time_prev;
    lptim_timing_stored_time_prev = hardware_time;
    lptim_timing_system_time += delta;
    CRITICAL_SECTION_END( );
}

static uint32_t lptim_timing_get_system_time( void )
{
    uint32_t current;
    current = lptim_get_hardware_time( );
    lptim_timing_update_system_time( current );

    return lptim_timing_system_time;
}
