/*!
 * @file      smtc_hal_lp_timer.h
 *
 * @brief     Board specific package RTC API definition.
 *
 * The Clear BSD License
 * Copyright Semtech Corporation 2023. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted (subject to the limitations in the disclaimer
 * below) provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Semtech corporation nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
 * THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL SEMTECH CORPORATION BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef SMTC_HAL_LP_TIMER_H
#define SMTC_HAL_LP_TIMER_H

#ifdef __cplusplus
extern "C" {
#endif

/*
 * -----------------------------------------------------------------------------
 * --- DEPENDENCIES ------------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PUBLIC MACROS -----------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PUBLIC CONSTANTS --------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PUBLIC TYPES ------------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PUBLIC FUNCTIONS PROTOTYPES ---------------------------------------------
 */

/*!
 *  @brief Initializes the MCU LPTIM32 peripheral
 */
void hal_lp_timer_init( void );

/*!
 * @brief Returns the current timer in seconds
 *
 * @remark Used for scheduling autonomous retransmissions (i.e: NbTrans),
 *         transmitting MAC answers, basically any delay without accurate time
 *         constraints. It is also used to measure the time spent inside the
 *         LoRaWAN process for the integrated failsafe.
 *
 * @returns rtc_time_s Current time in seconds
 */
uint32_t hal_lp_timer_get_time_s( void );

/*!
 * @brief Returns the current timer in milliseconds
 *
 * @remark Used to timestamp radio events (i.e: end of TX), will also be used
 * for ClassB
 *
 * @returns rtc_time_ms Current time in milliseconds
 */
uint32_t hal_lp_timer_get_time_ms( void );

/*!
 * @brief Returns the current timer in 0.1milliseconds
 *
 * @remark will also be used for D2D
 *
 * @returns rtc_time_100us Current time in 0.1ms
 */
uint32_t hal_lp_timer_get_time_100us( void );

/*!
 * @brief Sets the lptim wakeup timer for milliseconds parameter. The LPTIM will generate
 * an IRQ to wakeup the MCU.
 *
 * @param [in] milliseconds Number of seconds before wakeup
 */
void hal_lptim_wakeup_timer_set_ms( const int32_t milliseconds );

/*!
 * Return true if the wake up timer irq has been triggered
 *
 * \retval true if wut irq happended
 */
bool hal_lptim_has_wut_irq_happened( void );

/*!
 * @brief Stops the Timer
 */
void hal_lptim_wakeup_timer_stop( void );

#ifdef __cplusplus
}
#endif

#endif  // SMTC_HAL_LP_TIMER_H

/* --- EOF ------------------------------------------------------------------ */
