/*!
 * \file      smtc_hal_mcu.c
 *
 * \brief     MCU Hardware Abstraction Layer implementation
 *
 * The Clear BSD License
 * Copyright Semtech Corporation 2021. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted (subject to the limitations in the disclaimer
 * below) provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Semtech corporation nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
 * THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL SEMTECH CORPORATION BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * -----------------------------------------------------------------------------
 * --- DEPENDENCIES ------------------------------------------------------------
 */

#include <stdint.h>   // C99 types
#include <stdbool.h>  // bool type

#include "smtc_hal_mcu.h"
#include "modem_pinout.h"

#include "system_FM33A0XXEV.h"
#include "fm33a0xxev_include_all.h"

#include "smtc_hal.h"

#if( MODEM_HAL_DBG_TRACE == MODEM_HAL_FEATURE_ON )
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#endif
/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE MACROS-----------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE CONSTANTS -------------------------------------------------------
 */

#define RCLP_TRIM ( uint8_t )( *( uint32_t* ) 0X1FFFFB20UL )

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE TYPES -----------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE VARIABLES -------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE FUNCTIONS DECLARATION -------------------------------------------
 */

static void hal_mcu_delay_tick( const uint32_t tick_num );
static void mcu_system_init( void );
static void Init_RCLP( void );
static void Init_SysTick( void );
static void Init_SysClk_Gen( void );
static void Init_RCHF( void );
static void Init_SysClk( void );

static void mcu_clk_set_pll( void );

static void mcu_gpio_init( void );
static void DEBUG_Init( void );

/*!
 * @brief Function runing the low power mode handler
 */
static void hal_mcu_sleep_handler( void );

/*
 * -----------------------------------------------------------------------------
 * --- PUBLIC FUNCTIONS DEFINITION ---------------------------------------------
 */

void hal_mcu_critical_section_begin( uint32_t* mask )
{
    *mask = __get_PRIMASK( );
    __disable_irq( );
}

void hal_mcu_critical_section_end( uint32_t* mask )
{
    __set_PRIMASK( *mask );
}

void hal_mcu_disable_irq( void )
{
    __disable_irq( );
}

void hal_mcu_enable_irq( void )
{
    __enable_irq( );
}

void hal_mcu_init( void )
{
    /* Initialize MCU HAL library */
    mcu_system_init( );
    mcu_clk_set_pll( );

    mcu_gpio_init( );  // Initialize GPIOs

#if( MODEM_HAL_DBG_TRACE == MODEM_HAL_FEATURE_ON )
    hal_uart_init( DEBUG_UART_TX, DEBUG_UART_RX );
#endif

    hal_lp_timer_init( );
    hal_bstim_init( );
    hal_spi_init( RADIO_SPI_ID, RADIO_SPI_MOSI, RADIO_SPI_MISO, RADIO_SPI_SCLK );

    /* Initialize ADC */
    hal_adc_init( );
}

void hal_mcu_delay_ms( uint32_t delay_ms )
{
    uint32_t clk_num;

    clk_num = ( __SYSTEM_CLOCK / 1000 );

    for( ; delay_ms > 0; delay_ms-- )
    {
        if( delay_ms % 6000 == 0 )
        {
            hal_watchdog_reload( );  // reload watchdog no more than watchdog period ( 8s )
        }
        hal_mcu_delay_tick( clk_num );
    }
}

void hal_mcu_reset( void )
{
    __disable_irq( );
    NVIC_SystemReset( );  // Restart system
}

void __attribute__( ( optimize( "O0" ) ) ) hal_mcu_wait_us( uint32_t microseconds )
{
    uint32_t clk_num;

    clk_num = ( __SYSTEM_CLOCK / 1000000 );

    for( ; microseconds > 0; microseconds-- )
    {
        hal_mcu_delay_tick( clk_num );
    }
}

void hal_mcu_set_sleep_for_ms( const int32_t milliseconds )
{
    //ִߴźʱ 8 ms.
    if( milliseconds <= 8 )
    {
        return;
    }

    hal_lptim_wakeup_timer_set_ms( milliseconds - 8 );
    hal_mcu_sleep_handler( );

    mcu_clk_set_pll( );  //*************Reconfig system CLK to 64Mhz
}

#ifdef USE_FULL_ASSERT
/*
 * Function Name  : assert_failed
 * Description    : Reports the name of the source file and the source line
 * number where the assert_param error has occurred. Input          : - file:
 * pointer to the source file name
 *                  - line: assert_param error line source number
 * Output         : None
 * Return         : None
 */
void assert_failed( uint8_t* file, uint32_t line )
{
    // User can add his own implementation to report the file name and line
    // number,
    // ex: printf("Wrong parameters value: file %s on line %lu\r\n", file, line)

    SMTC_HAL_TRACE_PRINTF( "Wrong parameters value: file %s on line %lu\r\n", ( const char* ) file, line );
    // Infinite loop
    while( 1 )
    {
    }
}
#endif

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE FUNCTIONS DEFINITION --------------------------------------------
 */

static void hal_mcu_delay_tick( const uint32_t tick_num )
{
    uint32_t last = SysTick->VAL;

    uint32_t ticks = tick_num * clkmode;

    if( ticks > 0xF00000 )
    {
        ticks = 0xF00000;
    }

    while( ( ( last - SysTick->VAL ) & 0xFFFFFFUL ) < ticks )
        ;
}

static void mcu_system_init( void )
{
    __disable_irq( );
#if( HAL_USE_WATCHDOG == HAL_FEATURE_ON )
    hal_watchdog_init( );
    hal_watchdog_reload( );
#endif  // HAL_USE_WATCHDOG == HAL_FEATURE_ON
    Init_SysTick( );
    hal_mcu_wait_us( 10000 );
    Init_SysClk_Gen( );
    RMU_PDRCR_PDREN_Setable( ENABLE );  // Open PDR
    RMU_BORCR_BOR_PDRCFG_Set( RMU_BORCR_BOR_PDRCFG_1P75V );
    RMU_BORCR_OFF_BOR_Setable( DISABLE );  // Open BOR
    DEBUG_Init( );
    hal_mcu_wait_us( 10000 );
    __enable_irq( );
}

static void Init_RCLP( void )
{
    CDIF->CR = 0X0A;
    VRTC->RCLPTR = RCLP_TRIM;
    CDIF->CR = 0X05;
}

static void Init_SysTick( void )
{
    CMU->SYSCLKCR &= 0xFFFFFF3F;
    SysTick_Config( 0x1000000UL );  // 1ms tick
    SysTick->CTRL = SysTick_CTRL_CLKSOURCE_Msk | SysTick_CTRL_ENABLE_Msk;
}

static void Init_SysClk_Gen( void )
{
    Init_RCHF( );
    Init_RCLP();
    Init_SysClk( );
}

static void Init_RCHF( void )
{
    CMU_RCHF_InitTypeDef RCHF_InitStruct;

    RCHF_InitStruct.FSEL   = SYSCLKdef;
    RCHF_InitStruct.RCHFEN = ENABLE;

    CMU_RCHF_Init( &RCHF_InitStruct );

    CMU_Init_RCHF_Trim( clkmode );
}

static void Init_SysClk( void )
{
    CMU_SYSCLK_InitTypeDef SYSCLK_InitStruct;

    SYSCLK_InitStruct.SYSCLKSEL  = CMU_SYSCLKCR_SYSCLKSEL_RCHF;
    SYSCLK_InitStruct.AHBPRES    = CMU_SYSCLKCR_AHBPRES_DIV1;
    SYSCLK_InitStruct.APBPRES    = CMU_SYSCLKCR_APBPRES_DIV1;
    SYSCLK_InitStruct.SLP_ENEXTI = ENABLE;
    CMU_SysClk_Init( &SYSCLK_InitStruct );
}

static void mcu_clk_set_pll( void )
{
    uint32_t i = 0;

    CMU->PCLKCR1 |= 0x1 << 7;
    GPIOF->FCR |= 0x3C;

    CMU_XTHFCR_XTHFEN_Setable( ENABLE );
    CMU_XTHFCR_XTHF_CFG_Set( CMU_XTHFCR_XTHF_CFG_MAX );
    hal_mcu_delay_ms( 1 );

    FLS_RDCR_WAIT_Set( FLS_RDCR_WAIT_2CYCLE );
    hal_watchdog_reload( );

    CMU->PLLHCR |= 0x1 << 1;

    CMU_PLLHCR_REFPRSC_Set( CMU_PLLHCR_REFPRSC_DIV8 );

    CMU->PLLHCR &= ~( 0x3ff << 16 );
    CMU->PLLHCR |= 0x3f << 16;

    CMU_PLLHCR_EN_Setable( ENABLE );

    while( !CMU_PLLHCR_LOCKED_Chk( ) )
    {
        if( i >= 6400 )
        {
            break;
        }
        i++;
    }

    CMU_SYSCLKCR_SYSCLKSEL_Set( CMU_SYSCLKCR_SYSCLKSEL_PLL_H );
    CMU_SYSCLKCR_AHBPRES_Set( CMU_SYSCLKCR_AHBPRES_DIV1 );
    CMU_SYSCLKCR_APBPRES_Set( CMU_SYSCLKCR_APBPRES_DIV1 );
}

static void mcu_gpio_init( void )
{
    hal_gpio_init_out( RADIO_NSS, 1 );
    hal_gpio_init_in( RADIO_BUSY_PIN, BSP_GPIO_PULL_MODE_NONE, BSP_GPIO_IRQ_MODE_OFF, NULL );
    // Here init only the pin as an exti rising and the callback will be attached later
    hal_gpio_init_in( RADIO_DIOX, BSP_GPIO_PULL_MODE_DOWN, BSP_GPIO_IRQ_MODE_RISING, NULL );
    hal_gpio_init_out( RADIO_NRST, 1 );
#if defined( SX128X )
    hal_gpio_init_out( RADIO_ANTENNA_SWITCH, 1 );
#elif defined( LR11XX_TRANSCEIVER ) && defined( ENABLE_MODEM_GNSS_FEATURE )
    hal_gpio_init_out( RADIO_LNA_CTRL, 0 );
#elif defined( SX126X )
    // If the sx126x drives the rf switch with dio2, just put the SX126X_RADIO_RF_SWITCH_CTRL in pull up
//    hal_gpio_init_in( SX126X_RADIO_RF_SWITCH_CTRL, BSP_GPIO_PULL_MODE_UP, BSP_GPIO_IRQ_MODE_OFF, NULL );
    hal_gpio_init_out( SX126X_RADIO_RF_SWITCH_CTRL, 1 );
#endif
}

/**
 * @brief Low power handler
 *
 */
static void hal_mcu_sleep_handler( void )
{
    PMU_SleepCfg_InitTypeDef SleepCfg_InitStruct;
    CDIF_CR_INTF_EN_Setable( ENABLE );

    RMU_PDRCR_PDREN_Setable( ENABLE );  // enable PDR
    RMU_PDRCR_PDRCFG_Set( RMU_PDRCR_PDRCFG_1P5V );
    RMU_BORCR_OFF_BOR_Setable( ENABLE );
    VRTC_RCMFCR_EN_Setable( DISABLE );

    CDIF_CR_INTF_EN_Setable( DISABLE );

    SleepCfg_InitStruct.PMOD  = PMU_CR_PMOD_SLEEP;
    SleepCfg_InitStruct.SLPDP = PMU_CR_SLPDP_DEEPSLEEP;  // deepsleep
    SleepCfg_InitStruct.CVS   = DISABLE;
    SleepCfg_InitStruct.SCR   = 0;
    SleepCfg_InitStruct.TIA   = PMU_WKTR_T1A_8US;

    PMU_SleepCfg_Init( &SleepCfg_InitStruct );

    hal_watchdog_reload( );
    __WFI( );
    hal_watchdog_reload( );
    RMU_BORCR_OFF_BOR_Setable( DISABLE );  // enable BOR
}

static void DEBUG_Init( void )
{
    DBG->CR = 0x3;
}

/**
 * @brief This function handles Hard fault interrupt.
 */
void HardFault_Handler( void )
{
    SMTC_HAL_TRACE_ERROR( "\x1B[0;31m" );  // red color
    SMTC_HAL_TRACE_ERROR( "HARDFAULT_Handler\n" );
    SMTC_HAL_TRACE_ERROR( "\x1B[0m" );  // default color
    while( 1 )
    {
    }
}

/* --- EOF ------------------------------------------------------------------ */
