/*!
 * \file      smtc_hal_rng.c
 *
 * \brief     Random Number Generator Hardware Abstraction Layer implementation
 *
 * Revised BSD License
 * Copyright Semtech Corporation 2020. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Semtech corporation nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL SEMTECH CORPORATION BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * -----------------------------------------------------------------------------
 * --- DEPENDENCIES ------------------------------------------------------------
 */

#include <stdint.h>   // C99 types
#include <stdbool.h>  // bool type

#include "fm33a0xxev_rng.h"
#include "fm33a0xxev_cmu.h"
#include "smtc_hal_rng.h"
#include "smtc_hal_mcu.h"
#include "system_FM33A0XXEV.h"
#include "fm33a0xxev_include_all.h"

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE MACROS-----------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE CONSTANTS -------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE TYPES -----------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE VARIABLES -------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE FUNCTIONS DECLARATION -------------------------------------------
 */

static uint32_t RNG_ClkCalc( uint32_t RCHFClk );

/*
 * -----------------------------------------------------------------------------
 * --- PUBLIC FUNCTIONS DEFINITION ---------------------------------------------
 */
void hal_rng_init( void )
{
    CMU_ClocksType RCC_Clocks;
    uint32         RngClk;

    CMU_PERCLK_SetableEx( ANACCLK, ENABLE );
    CMU_PERCLK_SetableEx( TRNGCLK, ENABLE );

    /* һ㽨Ӧʹ4MʱΪʱ */
    CMU_GetClocksFreq( &RCC_Clocks );                   //ȡRCHFʱƵ
    RngClk = RNG_ClkCalc( RCC_Clocks.RCHF_Frequency );  //ȡʱӷƵ
    CMU_OPCCR2_RNGPRSC_Set( RngClk );                   //ʱӷƵ
    CMU_OPCCR2_RNGCKE_Setable( ENABLE );                //ʱʹ
}

uint32_t hal_rng_get_random( void )
{
    uint32_t rand_nb = 0;

    hal_rng_init( );

    RNG_SR_RNF_Clr( );
    RNG_CR_EN_Setable( ENABLE );
    hal_mcu_wait_us( 1000 );
    RNG_CR_EN_Setable( DISABLE );

    if( SET == RNG_SR_RNF_Chk( ) )
    {
        RNG_SR_RNF_Clr( );
        mcu_panic( );
    }

    rand_nb = RNG_DOR_Read( );
    return rand_nb;
}

uint32_t hal_rng_get_random_in_range( const uint32_t val_1, const uint32_t val_2 )
{
    if( val_1 <= val_2 )
    {
        return ( uint32_t )( ( hal_rng_get_random( ) % ( val_2 - val_1 + 1 ) ) + val_1 );
    }
    else
    {
        return ( uint32_t )( ( hal_rng_get_random( ) % ( val_1 - val_2 + 1 ) ) + val_2 );
    }
}

int32_t hal_rng_get_signed_random_in_range( const int32_t val_1, const int32_t val_2 )
{
    uint32_t tmp_range = 0;  // ( val_1 <= val_2 ) ? ( val_2 - val_1 ) : ( val_1 - val_2 );

    if( val_1 <= val_2 )
    {
        tmp_range = ( val_2 - val_1 );
        return ( int32_t )( ( val_1 + hal_rng_get_random_in_range( 0, tmp_range ) ) );
    }
    else
    {
        tmp_range = ( val_1 - val_2 );
        return ( int32_t )( ( val_2 + hal_rng_get_random_in_range( 0, tmp_range ) ) );
    }
}

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE FUNCTIONS DEFINITION --------------------------------------------
 */

/********************************
RNGʱӼ㺯
:RCHFClkRNGƵϵ̶4MHZ
룺RCHFClk
RNGʱӹ4MHZµķƵϵ
********************************/
static uint32_t RNG_ClkCalc( uint32_t RCHFClk )
{
    uint32_t div = 0;
    switch( RCHFClk )
    {
    case 16000000:
        div = CMU_OPCCR2_RNGPRSC_DIV4;
        break;
    case 24000000:
        div = CMU_OPCCR2_RNGPRSC_DIV8;
        break;
    case 8000000:
    default:
        div = CMU_OPCCR2_RNGPRSC_DIV2;
        break;
    }
    return div;
}

/* --- EOF ------------------------------------------------------------------ */
