/*!
 * \file      smtc_hal_spi.c
 *
 * \brief     SPI Hardware Abstraction Layer implementation
 *
 * The Clear BSD License
 * Copyright Semtech Corporation 2021. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted (subject to the limitations in the disclaimer
 * below) provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Semtech corporation nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
 * THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL SEMTECH CORPORATION BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * -----------------------------------------------------------------------------
 * --- DEPENDENCIES ------------------------------------------------------------
 */

#include <stdint.h>   // C99 types
#include <stdbool.h>  // bool type

#include "smtc_hal_spi.h"
#include "smtc_hal_trace.h"
#include "smtc_hal_mcu.h"

#include "system_FM33A0XXEV.h"
#include "fm33a0xxev_include_all.h"

#include "modem_pinout.h"

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE MACROS-----------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE CONSTANTS -------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE TYPES -----------------------------------------------------------
 */

typedef struct spi_s
{
    SPI_Type* interface;
    struct
    {
        hal_gpio_pin_names_t mosi;
        hal_gpio_pin_names_t miso;
        hal_gpio_pin_names_t sclk;
    } pins;
} spi_t;

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE VARIABLES -------------------------------------------------------
 */
static spi_t spi_periph[] = {
    [0] =
        {
            .interface = SPI3,
            .pins =
            {
                .mosi = NC,
                .miso = NC,
                .sclk = NC,
            },
        },
}
;

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE FUNCTIONS DECLARATION -------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PUBLIC FUNCTIONS DEFINITION ---------------------------------------------
 */

void hal_spi_init( const uint32_t id, const hal_gpio_pin_names_t mosi, const hal_gpio_pin_names_t miso,
                   const hal_gpio_pin_names_t sclk )
{
    // assert_param( ( id > 0 ) && ( ( id - 1 ) < sizeof( spi_periph ) ) );
    uint32_t local_id = 0;

    spi_periph[local_id].interface = SPI3;

    // Enable the clock
    CMU_PERCLK_SetableEx( SPI3CLK, ENABLE );  // Enable SPI3 Clock

    SPIx_CR1_IOSWAP_Set( SPI3, SPIx_CR1_IOSWAP_DEFAULT );  // MISO, MOSI not swap
    SPIx_CR1_MM_Set( SPI3, SPIx_CR1_MM_MASTER );           // master mode
    // SPIx_CR1_WAIT_Set(SPI3, SPIx_CR1_WAIT_1WAIT);                 //insert one clk after each one frame
    SPIx_CR1_BAUD_Set( SPI3, SPIx_CR1_BAUD_DIV8 );    // clock devide 64M/8
    SPIx_CR1_LSBF_Set( SPI3, SPIx_CR1_LSBF_MSB );     // MSB first
    SPIx_CR1_CPHOL_Set( SPI3, SPIx_CR1_CPHOL_LOW );   // CLK stop at low
    SPIx_CR1_CPHA_Set( SPI3, SPIx_CR1_CPHA_1CLOCK );  // Data sample at the first edge
    SPIx_CR2_SSNSEN_Setable( SPI3, DISABLE );         // NSS controled by hardware

    SPIx_CR2_RXO_Setable( SPI3, DISABLE );                     // full dumplex
    SPIx_CR2_DLEN_Set( SPI3, SPIx_CR2_DLEN_8BIT );             // 8 bit data
    SPIx_CR2_HALFDUPLEX_Set( SPI3, SPIx_CR2_HALFDUPLEX_SPI );  // Standard SPI, 4 line
    // SPIx_CR2_SSNM_Set( SPI3, SPIx_CR2_SSNM_LOW );              // NSS keep low after each transmit
    SPIx_CR2_TXO_AC_Setable( SPI3, DISABLE );
    SPIx_CR2_TXO_Setable( SPI3, DISABLE );

    // Clear the error flag
    SPIx_CR3_SERRC_Clr( SPI3 );
    SPIx_CR3_MERRC_Clr( SPI3 );
    // Clear the buffer
    SPIx_CR3_RXBFC_Clr( SPI3 );
    SPIx_CR3_TXBFC_Clr( SPI3 );

    SPIx_CR2_SPIEN_Setable( SPI3, ENABLE );  // Enable the SPI3

    spi_periph[local_id].pins.mosi = mosi;
    spi_periph[local_id].pins.miso = miso;
    spi_periph[local_id].pins.sclk = sclk;

    // Configure the SPI pins, and always SPI pins are within the same port
    GPIO_InitTypeDef GPIO_InitStructure;

    GPIO_Type* GPIOx         = ( GPIO_Type* ) ( GPIOA_BASE + ( mosi / 16 ) * 0x40 );
    uint32_t   GPIO_PIN_MOSI = ( uint32_t )( 1 << ( mosi & 0xF ) );
    uint32_t   GPIO_PIN_MISO = ( uint32_t )( 1 << ( miso & 0xF ) );
    uint32_t   GPIO_PIN_SCLK = ( uint32_t )( 1 << ( sclk & 0xF ) );

    GPIO_InitStructure.PxFCR  = GPIO_FCR_DIG;
    GPIO_InitStructure.PxINEN = GPIO_IN_Dis;
    GPIO_InitStructure.PxODEN = GPIO_OD_Dis;
    GPIO_InitStructure.PxPUEN = GPIO_PU_Dis;

    GPIO_InitStructure.Pin = GPIO_PIN_SCLK | GPIO_PIN_MISO | GPIO_PIN_MOSI;
    GPIO_Init( GPIOx, &GPIO_InitStructure );
}

void hal_spi_de_init( const uint32_t id )
{
    SPIx_CR2_SPIEN_Setable( SPI3, DISABLE );   // Disable the SPI3
    CMU_PERCLK_SetableEx( SPI3CLK, DISABLE );  // DISABLE SPI3 Clock
}

uint32_t hal_spi_in_out( const uint32_t id, uint32_t data )
{
    // assert_param( ( id > 0 ) && ( ( id - 1 ) < sizeof( spi_periph ) ) );
    uint32_t local_id = 0;

    while( !SPIx_ISR_TXBE_Chk( spi_periph[local_id].interface ) )
        ;

    SPIx_TXBUF_Write( spi_periph[local_id].interface, ( uint8_t )( data & 0xFF ) );

    while( !SPIx_ISR_TXBE_Chk( spi_periph[local_id].interface ) )
        ;

    while( !SPIx_ISR_RXBF_Chk( spi_periph[local_id].interface ) )
        ;

    return SPIx_RXBUF_Read( spi_periph[local_id].interface );
}

/* --- EOF ------------------------------------------------------------------ */
