/*!
 * \file      smtc_hal_uart.c
 *
 * \brief     UART Hardware Abstraction Layer implementation
 *
 * The Clear BSD License
 * Copyright Semtech Corporation 2021. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted (subject to the limitations in the disclaimer
 * below) provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Semtech corporation nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
 * THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL SEMTECH CORPORATION BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * -----------------------------------------------------------------------------
 * --- DEPENDENCIES ------------------------------------------------------------
 */

#include <stdint.h>   // C99 types
#include <stdbool.h>  // bool type

#include "system_FM33A0XXEV.h"
#include "fm33a0xxev_include_all.h"

#include "smtc_hal_uart.h"

#include "modem_pinout.h"

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE MACROS-----------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE CONSTANTS -------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE TYPES -----------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE VARIABLES -------------------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE FUNCTIONS DECLARATION -------------------------------------------
 */

/*
 * -----------------------------------------------------------------------------
 * --- PUBLIC FUNCTIONS DEFINITION ---------------------------------------------
 */

void hal_uart_init( const hal_gpio_pin_names_t uart_tx, const hal_gpio_pin_names_t uart_rx )
{
    UART_SInitTypeDef UART_para;
    CMU_ClocksType    CMU_Clocks;
    UART_para.ClockSrc = CMU_OPCCR1_UART1CKS_APBCLK;

    GPIO_InitTypeDef GPIO_InitStructure;

    GPIO_Type* GPIOx          = ( GPIO_Type* ) ( GPIOA_BASE + ( uart_tx / 16 ) * 0x40 );
    uint32_t   GPIO_PIN_TX    = ( uint32_t )( 1 << ( uart_tx & 0x0F ) );
    uint32_t   GPIO_PIN_RX    = ( uint32_t )( 1 << ( uart_rx & 0x0F ) );
    GPIO_InitStructure.PxFCR  = GPIO_FCR_DIG;
    GPIO_InitStructure.PxINEN = GPIO_IN_Dis;
    GPIO_InitStructure.PxODEN = GPIO_OD_Dis;
    GPIO_InitStructure.PxPUEN = GPIO_PU_Dis;
    GPIO_InitStructure.Pin    = GPIO_PIN_TX | GPIO_PIN_RX;
    GPIO_Init( GPIOx, &GPIO_InitStructure );

    // NVIC configuration
    NVIC_DisableIRQ( UART1_IRQn );

    // Baudrate configuration
    UART_para.BaudRate  = 115200;  // 921600
    UART_para.DataBit   = Eight8Bit;
    UART_para.ParityBit = NONE;
    UART_para.StopBit   = OneBit;

    CMU_GetClocksFreq( &CMU_Clocks );
    UART_SInit( UART1, &UART_para, &CMU_Clocks );
}

void hal_uart_deinit( void )
{
    UARTx_CSR_RXEN_Setable( UART1, DISABLE );  //رսʹ
    UARTx_CSR_TXEN_Setable( UART1, DISABLE );  //رշʹ

    CMU_OPCCR1_UART1CKE_Setable( DISABLE );     // UART1ʱӹر
    CMU_PERCLK_SetableEx( UART1CLK, DISABLE );  // UART1ʱӹر
}

void hal_uart_tx( uint8_t* buff, uint16_t len )
{
    CMU_PERCLK_SetableEx( DMACLK, ENABLE );

    DMA_GCR_DMAEN_Setable( ENABLE );
    DMA_CHxCR_CHxPRI_Set( DMA_CH6, DMA_CHxCR_CHxPRI_VERY_HIGH );
    DMA_CHxCR_CHxINC_Set( DMA_CH6, DMA_CHxCR_CHxINC_INCREASE );
    DMA_CHxCR_CHxSSEL_Set( DMA_CH6, DMA_CHxCR_CH6SSEL_UART1_TX );
    DMA_CHxCR_DIR_Set( DMA_CH6, DMA_CHxCR_DIR_TO_PER );
    DMA_CHxCR_CHxFTIE_Setable( DMA_CH6, DISABLE );
    DMA_CHxCR_CHxTSIZE_Set( DMA_CH6, len - 1 );
    DMA_CHxMAR_Write( DMA_CH6, ( uint32_t ) buff );

    UARTx_CSR_DMATXIFCFG_Setable( UART1, ENABLE );
    DMA_CHxCR_ChxEN_Setable( DMA_CH6, ENABLE );

    UARTx_CSR_TXEN_Setable( UART1, ENABLE );

    while( DMA_ISR_DMACHFT_Chk( DMA_CH6 ) == RESET )
    {
    };

    DMA_ISR_DMACHFT_Clr( DMA_CH6 );

    UARTx_ISR_TXSE_Clr( UART1 );

    while( UARTx_ISR_TXSE_Chk( UART1 ) == RESET )
    {
    };

    UARTx_ISR_TXSE_Clr( UART1 );
}

void hal_uart_rx( uint8_t* rx_buffer, uint16_t len )
{
    CMU_PERCLK_SetableEx( DMACLK, ENABLE );

    CMU_PERCLK_SetableEx( DMACLK, ENABLE );                        // DMAʱ
    DMA_GCR_DMAEN_Setable( ENABLE );                               // DMAʹ
    DMA_CHxCR_CHxPRI_Set( DMA_CH5, DMA_CHxCR_CHxPRI_HIGH );        //ͨȼ
    DMA_CHxCR_CHxINC_Set( DMA_CH5, DMA_CHxCR_CHxINC_INCREASE );    //ַͨ
    DMA_CHxCR_CHxSSEL_Set( DMA_CH5, DMA_CHxCR_CH5SSEL_UART1_RX );  //ͨѡ
    DMA_CHxCR_DIR_Set( DMA_CH5, DMA_CHxCR_DIR_TO_RAM );            //ݴ䷽
    DMA_CHxCR_CHxFTIE_Setable( DMA_CH5, ENABLE );                  //ͨжʹ
    DMA_CHxCR_CHxTSIZE_Set( DMA_CH5, len - 1 );
    DMA_CHxMAR_Write( DMA_CH5, ( uint32 ) rx_buffer );
    DMA_CHxCR_ChxEN_Setable( DMA_CH5, ENABLE );  //ͨ5ʹ

    UARTx_CSR_RXEN_Setable( UART1, ENABLE );  //ʹ

    while( DMA_ISR_DMACHFT_Chk( DMA_CH5 ) == RESET )
    {
    };
    DMA_ISR_DMACHFT_Clr( DMA_CH5 );  //DMAɱ־
}

/*
 * -----------------------------------------------------------------------------
 * --- PRIVATE FUNCTIONS DEFINITION --------------------------------------------
 */

/* --- EOF ------------------------------------------------------------------ */
